"""Contains constants used as part of plotting modules.

Copyright © 2023 Pixelgen Technologies AB.
"""


class Color:
    """Class for color constants."""

    LIGHTGREY = "#bebebe"
    ORANGERED2 = "#ee4000"
    SKYBLUE3 = "#6ca6cd"
    NETWORKX_NODE_COLOR = "#1f78b4"
    JETSET = [
        "#00007F",  # dark blue
        "#0000FF",  # blue
        "#007FFF",  # azure
        "#00FFFF",  # cyan
        "#7FFF7F",  # light green
        "#FFFF00",  # yellow
        "#FF7F00",  # orange
        "#FF0000",  # red
        "#7F0000",  # dark red
    ]
