"""MPX configuration.

Copyright © 2025 Pixelgen Technologies AB.
"""

from pixelator.common.config import get_position_in_parent, load_antibody_panel
from pixelator.mpx.config.config_instance import config

__all__ = [
    "config",
    "get_position_in_parent",
    "load_antibody_panel",
]
