"""Top level module for analysis subpackage of pixelator.pna.

Copyright © 2024 Pixelgen Technologies AB.
"""

from pixelator.pna.analysis.proximity import calculate_differential_proximity

__all__ = ["calculate_differential_proximity"]
