"""Model for report data returned by the single-cell layout stage.

Copyright © 2024 Pixelgen Technologies AB.
"""

from __future__ import annotations

from pixelator.pna.report.models.base import SampleReport


class LayoutSampleReport(SampleReport):
    """Model for report data returned by the single-cell layout stage."""

    pass
