"""Constants for the collapse module.

This module contains constants for the collapse and error correction
module of MPX data (from FASTQ).

Copyright © 2023 Pixelgen Technologies AB.
"""

SEED = 42
