"""Constants used as part of the annotation modules.

Copyright © 2022 Pixelgen Technologies AB.
"""

MINIMUM_NBR_OF_CELLS_FOR_ANNOTATION = 20

# size limit constants
MINIMUM_N_EDGES_CELL_SIZE = 300
MINIMUM_NBR_OF_CELLS_FOR_SIZE_LIMIT = 20
CELL_MIN_SIZE_SMOOTHING_FACTOR = 0.8
CELL_MAX_SIZE_SMOOTHING_FACTOR = 0
DISTANCE_DEVIATION_FACTOR = 5
# this is number of edges
PRE_FILTER_LIMIT = 20

TAU_HARD_THRESHOLD = 0.995
TAU_IQR_UPPER_THRESHOLD = 2
TAU_IQR_LOWER_THRESHOLD = 5
