"""Keep constants used on graph operations in pixelator.

Copyright © 2023 Pixelgen Technologies AB.
"""

MIN_PIXELS_TO_REFINE = 100
LEIDEN_RESOLUTION = 0.01
RELATIVE_ANNOTATE_RESOLUTION = (
    0.5  # A lower resolution is used for annotation of potential doublets
)
